from .messages_pb2 import (
    AuthenticationRequestMessage,
    AuthenticationResponseMessage,
    OpenTunnelRequestMessage,
    OpenTunnelResponseMessage,
    PingMessage,
    PongMessage,
    RawRequestMessage,
    RawResponseMessage,
    MessageType,
    TunnelisMessage
)
from time import time as now


def build_authentication_http_request_message(user, domain, authtoken):
    """ Builds messages of type MSG_AUTHENTICATION_REQUEST for HTTP/HTTPS tunnels.
    """

    message = AuthenticationRequestMessage()
    message.user = user
    message.domain = domain
    message.authtoken = authtoken
    message.protocol = 'HTTP'

    tunnelis_message = TunnelisMessage()
    tunnelis_message.type = MessageType.MSG_AUTHENTICATION_REQUEST
    tunnelis_message.authentication_request.CopyFrom(message)

    return tunnelis_message


def build_authentication_tcp_request_message(user, domain, authtoken):
    """ Builds messages of type MSG_AUTHENTICATION_REQUEST for TCP tunnels.
    """

    message = AuthenticationRequestMessage()
    message.user = user
    message.domain = domain
    message.authtoken = authtoken
    message.protocol = 'TCP'

    tunnelis_message = TunnelisMessage()
    tunnelis_message.type = MessageType.MSG_AUTHENTICATION_REQUEST
    tunnelis_message.authentication_request.CopyFrom(message)

    return tunnelis_message


def build_authentication_ws_request_message(user, domain, authtoken):
    """ Builds messages of type MSG_AUTHENTICATION_REQUEST for WS tunnels.
    """

    message = AuthenticationRequestMessage()
    message.user = user
    message.domain = domain
    message.authtoken = authtoken
    message.protocol = 'WS'

    tunnelis_message = TunnelisMessage()
    tunnelis_message.type = MessageType.MSG_AUTHENTICATION_REQUEST
    tunnelis_message.authentication_request.CopyFrom(message)

    return tunnelis_message


def build_authentication_success_response_message(session_token):
    """ Builds messages of type MSG_AUTHENTICATION_RESPONSE.
    For cases when client authentication succeeds.
    """

    message = AuthenticationResponseMessage()
    message.session_token = session_token
    message.reason = ''
    message.success = True

    tunnelis_message = TunnelisMessage()
    tunnelis_message.type = MessageType.MSG_AUTHENTICATION_RESPONSE
    tunnelis_message.authentication_response.CopyFrom(message)

    return tunnelis_message


def build_authentication_failure_response_message(reason):
    """ Builds messages of type MSG_AUTHENTICATION_RESPONSE.
    For cases when client authentication fails.
    """

    message = AuthenticationResponseMessage()
    message.session_token = ''
    message.reason = reason
    message.success = False

    tunnelis_message = TunnelisMessage()
    tunnelis_message.type = MessageType.MSG_AUTHENTICATION_RESPONSE
    tunnelis_message.authentication_response.CopyFrom(message)

    return tunnelis_message


def build_open_tunnel_request(session_token):
    """
    Generic function to build messages of type MSG_OPEN_TUNNEL_REQUEST.
    """

    message = OpenTunnelRequestMessage()
    message.session_token = session_token

    tunnelis_message = TunnelisMessage()
    tunnelis_message.type = MessageType.MSG_OPEN_TUNNEL_REQUEST
    tunnelis_message.open_tunnel_request.CopyFrom(message)

    return tunnelis_message


def build_open_tunnel_success_response_message(url):
    """ Builds messages of type MSG_OPEN_TUNNEL_RESPONSE.
    For cases when tunnel openning succeeds.
    """

    message = OpenTunnelResponseMessage()
    message.url = url
    message.reason = ''
    message.success = True

    tunnelis_message = TunnelisMessage()
    tunnelis_message.type = MessageType.MSG_OPEN_TUNNEL_RESPONSE
    tunnelis_message.open_tunnel_response.CopyFrom(message)

    return tunnelis_message


def build_open_tunnel_failure_response_message(reason):
    """ Builds messages of type MSG_OPEN_TUNNEL_RESPONSE.
    For cases when tunnel openning fails.
    """

    message = OpenTunnelResponseMessage()
    message.url = ''
    message.reason = reason
    message.success = False

    tunnelis_message = TunnelisMessage()
    tunnelis_message.type = MessageType.MSG_OPEN_TUNNEL_RESPONSE
    tunnelis_message.open_tunnel_response.CopyFrom(message)

    return tunnelis_message


def build_ping_message():
    """ Builds messages of type MSG_PING.
    """

    message = PingMessage()

    tunnelis_message = TunnelisMessage()
    tunnelis_message.type = MessageType.MSG_PING
    tunnelis_message.ping_message.CopyFrom(message)

    return tunnelis_message


def build_pong_message(session_token):
    """ Builds messages of type MSG_PONG.
    """

    message = PongMessage()
    message.session_token = session_token
    message.timestamp = now()

    tunnelis_message = TunnelisMessage()
    tunnelis_message.type = MessageType.MSG_PONG
    tunnelis_message.pong_message.CopyFrom(message)

    return tunnelis_message


def build_raw_request_message(source_id, payload):
    """ Builds messages of type MSG_RAW_REQUEST.
    """

    message = RawRequestMessage()
    message.source_id = source_id
    message.payload = payload

    tunnelis_message = TunnelisMessage()
    tunnelis_message.type = MessageType.MSG_RAW_REQUEST
    tunnelis_message.raw_request.CopyFrom(message)

    return tunnelis_message


def build_raw_response_message(session_token, destination_id, payload):
    """ Builds messages of type MSG_RAW_RESPONSE.
    """

    message = RawResponseMessage()
    message.session_token = session_token
    message.destination_id = destination_id
    message.payload = payload

    tunnelis_message = TunnelisMessage()
    tunnelis_message.type = MessageType.MSG_RAW_RESPONSE
    tunnelis_message.raw_response.CopyFrom(message)

    return tunnelis_message



from typing import Any, Callable, Tuple
from zmq import DEALER as ZMQ_DEALER
from zmq.auth import load_certificate
from zmq.eventloop.zmqstream import ZMQStream
from zmq.sugar.context import Context


TCPMessageCallback = Callable[[Any, bytes], None]


class Client:
    """
    A configurable TCP client.

    A wrapper around ZMQ sockets of type ZMQ_DEALER.

    """

    __message_handler: TCPMessageCallback = None

    def __init__(
        self,
        secure: bool = False,
        client_private_key: str = 'client.key_secret',
        server_public_key: str = 'server.key'
    ):
        """
        Initializes client socket.

        Parameters
        ----------
        server_address : Tuple(str, int)
            The network address (host, port) of the server to connect to.

        secure : boolean
            Whether or not to secure communications.
            Communications are always secured using CURVE.

        client_private_key : str
            The complete path for this client's private key.
            Only relevant when using secured communications.

        server_public_key : str
            The complete path for the server's public key.
            Only relevant when using secured communications.

        """

        self.__zmq_context = Context()

        self.__socket = self.__zmq_context.socket(ZMQ_DEALER)

        if secure:

            server_public, _ = load_certificate(server_public_key)
            client_public, client_secret = load_certificate(client_private_key)

            self.__socket.curve_secretkey = client_secret
            self.__socket.curve_publickey = client_public
            self.__socket.curve_serverkey = server_public

        self.stream = ZMQStream(self.__socket)
        self.stream.on_recv(self.__message_handler)

    def connect(self, address: Tuple[str, int]) -> None:
        """
        Connect to a TCP address.

        Parameters
        ----------
        address : Tuple(str, int)
            The TCP network address to which to connect to.

        """
        host, port = address
        self.__socket.connect(f'tcp://{host}:{port}')

    def send(self, message: bytes) -> None:
        """
        Send raw application data to a remote client.

        Parameters
        ----------
        message : bytes
            The TCP message to be sent.

        """
        self.stream.send_multipart([message])

    def close(self) -> None:
        """
        Close endpoint.

        """
        self.stream.close()
        self.__socket.close()
        self.__zmq_context.term()

    def set_message_handler(self, handler: TCPMessageCallback) -> None:
        """
        Set the callback function to handle incoming TCP messages.

        Parameters
        ----------
        handler : Callable(Client, bytes) -> None
            The callback function.

        """
        self.__message_handler = handler

    def unset_message_handler(self) -> None:
        """
        Stop handling incoming TCP messages.

        """
        self.__message_handler = None

    def __message_handler(self, message: Tuple[bytes, bytes]) -> None:
        """
        A generic message handler.
        Forwards received messages to their proper message handlers.

        Parameters
        ----------
        message : Tuple(bytes, bytes)
            The received message.

        """

        if self.__message_handler:
            self.__message_handler(self, message[0])



import random
import string


def generate_id(n_chars=32):
    return ''.join(
        random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits + '_') for _ in range(n_chars)
    )


__all__ = ["generate_id"]



from typing import Any, Callable, Tuple
from zmq import ROUTER as ZMQ_ROUTER
from zmq.auth import load_certificate
from zmq.eventloop.zmqstream import ZMQStream
from zmq.sugar.context import Context


TCPMessageCallback = Callable[[Any, bytes, bytes], None]


class Server:
    """
    A configurable TCP server.

    A wrapper around ZMQ sockets of type ZMQ_ROUTER.

    """

    __message_handler: TCPMessageCallback = None

    def __init__(
        self,
        secure: bool = False,
        private_key: str = 'server.key_secret'
    ):
        """
        Initializes server socket.

        Parameters
        ----------
        secure : boolean
            Whether or not to secure communications.
            Communications are always secured using CURVE.

        private_key : str
            The complete path for this server's private key.
            Only relevant when using secured communications.

        """

        self.__zmq_context = Context()

        self.__socket = self.__zmq_context.socket(ZMQ_ROUTER)

        if secure:

            # Create endpoint for clients to communicate with server
            server_public, server_secret = load_certificate(private_key)

            self.__socket.curve_secretkey = server_secret
            self.__socket.curve_publickey = server_public
            self.__socket.curve_server = True

        self.stream = ZMQStream(self.__socket)
        self.stream.on_recv(self.__message_handler)

    def bind(self, port: int) -> None:
        """
        Bind socket to a TCP address and start listening for communications.

        Parameters
        ----------
        port : int
            The network port at which to bind the socket.
            When set to a non-positive integer the socket will bind to a random free port.

        """

        if port > 0:
            self.__socket.bind(f'tcp://*:{port}')
            self.port = port
        else:
            self.port = self.__socket.bind_to_random_port('tcp://*', min_port=30000)

    def send(self, identity: bytes, message: bytes) -> None:
        """
        Send raw application data to a remote client.

        Parameters
        ----------
        identity : bytes
            The identity (ZMQ_IDENTITY) of the remote client.

        message : bytes
            The TCP message to be sent.

        """
        self.stream.send_multipart([identity, message])

    def close(self) -> None:
        """
        Close endpoint.

        """
        self.stream.close()
        self.__socket.close()
        self.__zmq_context.term()

    def set_message_handler(self, handler: TCPMessageCallback) -> None:
        """
        Set the callback function to handle incoming TCP messages.

        Parameters
        ----------
        handler : Callable(Server, bytes, bytes) -> None
            The callback function.

        """
        self.__message_handler = handler

    def unset_message_handler(self) -> None:
        """
        Stop handling incoming TCP messages.

        """
        self.__message_handler = None

    def __message_handler(self, message: Tuple[bytes, bytes]) -> None:
        """
        A generic message handler.
        Forwards received messages to their proper message handlers.

        Parameters
        ----------
        message : Tuple(bytes, bytes)
            The received message.

        """

        identity, payload = message

        if self.__message_handler:
            self.__message_handler(self, identity, payload)



from tunnelis_core.sockets.client import *
from tunnelis_core.sockets.publisher import *
from tunnelis_core.sockets.server import *
from tunnelis_core.sockets.subscriber import *
from tunnelis_core.sockets.tcp import *


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: messages.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='messages.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0emessages.proto\"a\n\x1c\x41uthenticationRequestMessage\x12\x0c\n\x04user\x18\x01 \x01(\t\x12\x0e\n\x06\x64omain\x18\x02 \x01(\t\x12\x11\n\tauthtoken\x18\x03 \x01(\t\x12\x10\n\x08protocol\x18\x04 \x01(\t\"W\n\x1d\x41uthenticationResponseMessage\x12\x15\n\rsession_token\x18\x01 \x01(\t\x12\x0e\n\x06reason\x18\x02 \x01(\t\x12\x0f\n\x07success\x18\x03 \x01(\x08\"1\n\x18OpenTunnelRequestMessage\x12\x15\n\rsession_token\x18\x01 \x01(\t\"I\n\x19OpenTunnelResponseMessage\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x0e\n\x06reason\x18\x02 \x01(\t\x12\x0f\n\x07success\x18\x03 \x01(\x08\"\r\n\x0bPingMessage\"7\n\x0bPongMessage\x12\x15\n\rsession_token\x18\x01 \x01(\t\x12\x11\n\ttimestamp\x18\x02 \x01(\x01\"7\n\x11RawRequestMessage\x12\x11\n\tsource_id\x18\x01 \x01(\x0c\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\"T\n\x12RawResponseMessage\x12\x15\n\rsession_token\x18\x01 \x01(\t\x12\x16\n\x0e\x64\x65stination_id\x18\x02 \x01(\x0c\x12\x0f\n\x07payload\x18\x03 \x01(\x0c\"\xd6\x03\n\x0fTunnelisMessage\x12\x1a\n\x04type\x18\x01 \x01(\x0e\x32\x0c.MessageType\x12?\n\x16\x61uthentication_request\x18\x02 \x01(\x0b\x32\x1d.AuthenticationRequestMessageH\x00\x12\x41\n\x17\x61uthentication_response\x18\x03 \x01(\x0b\x32\x1e.AuthenticationResponseMessageH\x00\x12\x38\n\x13open_tunnel_request\x18\x04 \x01(\x0b\x32\x19.OpenTunnelRequestMessageH\x00\x12:\n\x14open_tunnel_response\x18\x05 \x01(\x0b\x32\x1a.OpenTunnelResponseMessageH\x00\x12$\n\x0cping_message\x18\x06 \x01(\x0b\x32\x0c.PingMessageH\x00\x12$\n\x0cpong_message\x18\x07 \x01(\x0b\x32\x0c.PongMessageH\x00\x12)\n\x0braw_request\x18\x08 \x01(\x0b\x32\x12.RawRequestMessageH\x00\x12+\n\x0craw_response\x18\t \x01(\x0b\x32\x13.RawResponseMessageH\x00\x42\t\n\x07message*\xe1\x01\n\x0bMessageType\x12\x0f\n\x0bMSG_UNKNOWN\x10\x00\x12\x1e\n\x1aMSG_AUTHENTICATION_REQUEST\x10\x01\x12\x1f\n\x1bMSG_AUTHENTICATION_RESPONSE\x10\x02\x12\x1b\n\x17MSG_OPEN_TUNNEL_REQUEST\x10\x03\x12\x1c\n\x18MSG_OPEN_TUNNEL_RESPONSE\x10\x04\x12\x0c\n\x08MSG_PING\x10\x05\x12\x0c\n\x08MSG_PONG\x10\x06\x12\x13\n\x0fMSG_RAW_REQUEST\x10\x07\x12\x14\n\x10MSG_RAW_RESPONSE\x10\x08\x62\x06proto3'
)

_MESSAGETYPE = _descriptor.EnumDescriptor(
  name='MessageType',
  full_name='MessageType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MSG_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MSG_AUTHENTICATION_REQUEST', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MSG_AUTHENTICATION_RESPONSE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MSG_OPEN_TUNNEL_REQUEST', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MSG_OPEN_TUNNEL_RESPONSE', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MSG_PING', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MSG_PONG', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MSG_RAW_REQUEST', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MSG_RAW_RESPONSE', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1021,
  serialized_end=1246,
)
_sym_db.RegisterEnumDescriptor(_MESSAGETYPE)

MessageType = enum_type_wrapper.EnumTypeWrapper(_MESSAGETYPE)
MSG_UNKNOWN = 0
MSG_AUTHENTICATION_REQUEST = 1
MSG_AUTHENTICATION_RESPONSE = 2
MSG_OPEN_TUNNEL_REQUEST = 3
MSG_OPEN_TUNNEL_RESPONSE = 4
MSG_PING = 5
MSG_PONG = 6
MSG_RAW_REQUEST = 7
MSG_RAW_RESPONSE = 8



_AUTHENTICATIONREQUESTMESSAGE = _descriptor.Descriptor(
  name='AuthenticationRequestMessage',
  full_name='AuthenticationRequestMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='AuthenticationRequestMessage.user', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='domain', full_name='AuthenticationRequestMessage.domain', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='authtoken', full_name='AuthenticationRequestMessage.authtoken', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='protocol', full_name='AuthenticationRequestMessage.protocol', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18,
  serialized_end=115,
)


_AUTHENTICATIONRESPONSEMESSAGE = _descriptor.Descriptor(
  name='AuthenticationResponseMessage',
  full_name='AuthenticationResponseMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='session_token', full_name='AuthenticationResponseMessage.session_token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reason', full_name='AuthenticationResponseMessage.reason', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='success', full_name='AuthenticationResponseMessage.success', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=117,
  serialized_end=204,
)


_OPENTUNNELREQUESTMESSAGE = _descriptor.Descriptor(
  name='OpenTunnelRequestMessage',
  full_name='OpenTunnelRequestMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='session_token', full_name='OpenTunnelRequestMessage.session_token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=206,
  serialized_end=255,
)


_OPENTUNNELRESPONSEMESSAGE = _descriptor.Descriptor(
  name='OpenTunnelResponseMessage',
  full_name='OpenTunnelResponseMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='OpenTunnelResponseMessage.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reason', full_name='OpenTunnelResponseMessage.reason', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='success', full_name='OpenTunnelResponseMessage.success', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=257,
  serialized_end=330,
)


_PINGMESSAGE = _descriptor.Descriptor(
  name='PingMessage',
  full_name='PingMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=332,
  serialized_end=345,
)


_PONGMESSAGE = _descriptor.Descriptor(
  name='PongMessage',
  full_name='PongMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='session_token', full_name='PongMessage.session_token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='PongMessage.timestamp', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=347,
  serialized_end=402,
)


_RAWREQUESTMESSAGE = _descriptor.Descriptor(
  name='RawRequestMessage',
  full_name='RawRequestMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='source_id', full_name='RawRequestMessage.source_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload', full_name='RawRequestMessage.payload', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=404,
  serialized_end=459,
)


_RAWRESPONSEMESSAGE = _descriptor.Descriptor(
  name='RawResponseMessage',
  full_name='RawResponseMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='session_token', full_name='RawResponseMessage.session_token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='destination_id', full_name='RawResponseMessage.destination_id', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payload', full_name='RawResponseMessage.payload', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=461,
  serialized_end=545,
)


_TUNNELISMESSAGE = _descriptor.Descriptor(
  name='TunnelisMessage',
  full_name='TunnelisMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='TunnelisMessage.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='authentication_request', full_name='TunnelisMessage.authentication_request', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='authentication_response', full_name='TunnelisMessage.authentication_response', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='open_tunnel_request', full_name='TunnelisMessage.open_tunnel_request', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='open_tunnel_response', full_name='TunnelisMessage.open_tunnel_response', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ping_message', full_name='TunnelisMessage.ping_message', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='pong_message', full_name='TunnelisMessage.pong_message', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='raw_request', full_name='TunnelisMessage.raw_request', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='raw_response', full_name='TunnelisMessage.raw_response', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='message', full_name='TunnelisMessage.message',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=548,
  serialized_end=1018,
)

_TUNNELISMESSAGE.fields_by_name['type'].enum_type = _MESSAGETYPE
_TUNNELISMESSAGE.fields_by_name['authentication_request'].message_type = _AUTHENTICATIONREQUESTMESSAGE
_TUNNELISMESSAGE.fields_by_name['authentication_response'].message_type = _AUTHENTICATIONRESPONSEMESSAGE
_TUNNELISMESSAGE.fields_by_name['open_tunnel_request'].message_type = _OPENTUNNELREQUESTMESSAGE
_TUNNELISMESSAGE.fields_by_name['open_tunnel_response'].message_type = _OPENTUNNELRESPONSEMESSAGE
_TUNNELISMESSAGE.fields_by_name['ping_message'].message_type = _PINGMESSAGE
_TUNNELISMESSAGE.fields_by_name['pong_message'].message_type = _PONGMESSAGE
_TUNNELISMESSAGE.fields_by_name['raw_request'].message_type = _RAWREQUESTMESSAGE
_TUNNELISMESSAGE.fields_by_name['raw_response'].message_type = _RAWRESPONSEMESSAGE
_TUNNELISMESSAGE.oneofs_by_name['message'].fields.append(
  _TUNNELISMESSAGE.fields_by_name['authentication_request'])
_TUNNELISMESSAGE.fields_by_name['authentication_request'].containing_oneof = _TUNNELISMESSAGE.oneofs_by_name['message']
_TUNNELISMESSAGE.oneofs_by_name['message'].fields.append(
  _TUNNELISMESSAGE.fields_by_name['authentication_response'])
_TUNNELISMESSAGE.fields_by_name['authentication_response'].containing_oneof = _TUNNELISMESSAGE.oneofs_by_name['message']
_TUNNELISMESSAGE.oneofs_by_name['message'].fields.append(
  _TUNNELISMESSAGE.fields_by_name['open_tunnel_request'])
_TUNNELISMESSAGE.fields_by_name['open_tunnel_request'].containing_oneof = _TUNNELISMESSAGE.oneofs_by_name['message']
_TUNNELISMESSAGE.oneofs_by_name['message'].fields.append(
  _TUNNELISMESSAGE.fields_by_name['open_tunnel_response'])
_TUNNELISMESSAGE.fields_by_name['open_tunnel_response'].containing_oneof = _TUNNELISMESSAGE.oneofs_by_name['message']
_TUNNELISMESSAGE.oneofs_by_name['message'].fields.append(
  _TUNNELISMESSAGE.fields_by_name['ping_message'])
_TUNNELISMESSAGE.fields_by_name['ping_message'].containing_oneof = _TUNNELISMESSAGE.oneofs_by_name['message']
_TUNNELISMESSAGE.oneofs_by_name['message'].fields.append(
  _TUNNELISMESSAGE.fields_by_name['pong_message'])
_TUNNELISMESSAGE.fields_by_name['pong_message'].containing_oneof = _TUNNELISMESSAGE.oneofs_by_name['message']
_TUNNELISMESSAGE.oneofs_by_name['message'].fields.append(
  _TUNNELISMESSAGE.fields_by_name['raw_request'])
_TUNNELISMESSAGE.fields_by_name['raw_request'].containing_oneof = _TUNNELISMESSAGE.oneofs_by_name['message']
_TUNNELISMESSAGE.oneofs_by_name['message'].fields.append(
  _TUNNELISMESSAGE.fields_by_name['raw_response'])
_TUNNELISMESSAGE.fields_by_name['raw_response'].containing_oneof = _TUNNELISMESSAGE.oneofs_by_name['message']
DESCRIPTOR.message_types_by_name['AuthenticationRequestMessage'] = _AUTHENTICATIONREQUESTMESSAGE
DESCRIPTOR.message_types_by_name['AuthenticationResponseMessage'] = _AUTHENTICATIONRESPONSEMESSAGE
DESCRIPTOR.message_types_by_name['OpenTunnelRequestMessage'] = _OPENTUNNELREQUESTMESSAGE
DESCRIPTOR.message_types_by_name['OpenTunnelResponseMessage'] = _OPENTUNNELRESPONSEMESSAGE
DESCRIPTOR.message_types_by_name['PingMessage'] = _PINGMESSAGE
DESCRIPTOR.message_types_by_name['PongMessage'] = _PONGMESSAGE
DESCRIPTOR.message_types_by_name['RawRequestMessage'] = _RAWREQUESTMESSAGE
DESCRIPTOR.message_types_by_name['RawResponseMessage'] = _RAWRESPONSEMESSAGE
DESCRIPTOR.message_types_by_name['TunnelisMessage'] = _TUNNELISMESSAGE
DESCRIPTOR.enum_types_by_name['MessageType'] = _MESSAGETYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AuthenticationRequestMessage = _reflection.GeneratedProtocolMessageType('AuthenticationRequestMessage', (_message.Message,), {
  'DESCRIPTOR' : _AUTHENTICATIONREQUESTMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:AuthenticationRequestMessage)
  })
_sym_db.RegisterMessage(AuthenticationRequestMessage)

AuthenticationResponseMessage = _reflection.GeneratedProtocolMessageType('AuthenticationResponseMessage', (_message.Message,), {
  'DESCRIPTOR' : _AUTHENTICATIONRESPONSEMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:AuthenticationResponseMessage)
  })
_sym_db.RegisterMessage(AuthenticationResponseMessage)

OpenTunnelRequestMessage = _reflection.GeneratedProtocolMessageType('OpenTunnelRequestMessage', (_message.Message,), {
  'DESCRIPTOR' : _OPENTUNNELREQUESTMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:OpenTunnelRequestMessage)
  })
_sym_db.RegisterMessage(OpenTunnelRequestMessage)

OpenTunnelResponseMessage = _reflection.GeneratedProtocolMessageType('OpenTunnelResponseMessage', (_message.Message,), {
  'DESCRIPTOR' : _OPENTUNNELRESPONSEMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:OpenTunnelResponseMessage)
  })
_sym_db.RegisterMessage(OpenTunnelResponseMessage)

PingMessage = _reflection.GeneratedProtocolMessageType('PingMessage', (_message.Message,), {
  'DESCRIPTOR' : _PINGMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:PingMessage)
  })
_sym_db.RegisterMessage(PingMessage)

PongMessage = _reflection.GeneratedProtocolMessageType('PongMessage', (_message.Message,), {
  'DESCRIPTOR' : _PONGMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:PongMessage)
  })
_sym_db.RegisterMessage(PongMessage)

RawRequestMessage = _reflection.GeneratedProtocolMessageType('RawRequestMessage', (_message.Message,), {
  'DESCRIPTOR' : _RAWREQUESTMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:RawRequestMessage)
  })
_sym_db.RegisterMessage(RawRequestMessage)

RawResponseMessage = _reflection.GeneratedProtocolMessageType('RawResponseMessage', (_message.Message,), {
  'DESCRIPTOR' : _RAWRESPONSEMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:RawResponseMessage)
  })
_sym_db.RegisterMessage(RawResponseMessage)

TunnelisMessage = _reflection.GeneratedProtocolMessageType('TunnelisMessage', (_message.Message,), {
  'DESCRIPTOR' : _TUNNELISMESSAGE,
  '__module__' : 'messages_pb2'
  # @@protoc_insertion_point(class_scope:TunnelisMessage)
  })
_sym_db.RegisterMessage(TunnelisMessage)


# @@protoc_insertion_point(module_scope)



from typing import Callable, Dict, Tuple
from zmq import SUB as ZMQ_SUB, SUBSCRIBE as ZMQ_SUBSCRIBE, UNSUBSCRIBE as ZMQ_UNSUBSCRIBE
from zmq.eventloop import zmqstream
from zmq.sugar.context import Context

SubscriberCallback = Callable[[bytes], None]


class Subscriber:
    """
    A generic message subscriber.

    A wrapper class around ZMQ sockets of type ZMQ_SUB.

    """

    __subscriptions: Dict[bytes, SubscriberCallback] = {}

    def __init__(self, broker: Tuple[str, int]):
        """
        Initialize a message subscriber.

        Parameters
        ----------
        broker : Tuple(str, int)
            The TCP network address (host, port) of the message broker.

        """

        broker_host, broker_port = broker

        self.__zmq_context = Context()

        # Connect subscriber to message broker
        self.__socket = self.__zmq_context.socket(ZMQ_SUB)
        self.__socket.connect(f'tcp://{broker_host}:{broker_port}')

        # Create stream to handle received messages asynchronously
        self.__stream = zmqstream.ZMQStream(self.__socket)
        self.__stream.on_recv(self.__message_handler)

    def subscribe(self, topic: bytes, handler: SubscriberCallback) -> None:
        """
        Registers a callback function and starts listening for messages published on a given topic.
        Only one handler function is allowed per topic.

        Parameters
        ----------
        topic : bytes
            The message topic.

        handler : Callable(bytes) -> None
            The callback function to call.

        """
        if topic not in self.__subscriptions:
            self.__socket.setsockopt(ZMQ_SUBSCRIBE, topic)
        self.__subscriptions[topic] = handler

    def unsubscribe(self, topic: bytes) -> None:
        """
        Stops listening for messages published on a given topic and removes registered callback function.

        Parameters
        ----------
        topic : bytes

        """
        if topic in self.__subscriptions:
            self.__socket.setsockopt(ZMQ_UNSUBSCRIBE, topic)
            del self.__subscriptions[topic]

    def close(self) -> None:
        """
        Stops subscriber.

        """
        self.__stream.close()
        self.__zmq_context.term()

    def __message_handler(self, message: Tuple[bytes, bytes]) -> None:
        """
        A generic message handler.
        Forwards received messages to their proper message handlers.

        Parameters
        ----------
        message : Tuple(bytes, bytes)
            The received message containing both the topic and the encoded event.

        """

        topic, event = message

        # NOTE: If a common handler exists for all topics then there's no need to
        # look for a specific topic handler.
        if b'' in self.__subscriptions:
            handler = self.__subscriptions[b'']
            handler(event)

        elif topic in self.__subscriptions:
            handler = self.__subscriptions[topic]
            handler(event)



import setuptools

setuptools.setup(
    name="tunnelis_core",
    version="1.6.4",
    author="Pedro Melo, Rafael Arrais, Sergio Marinho",
    author_email="pedro.m.melo@inesctec.pt, rafael.l.arrais@inesctec.pt, sergio.d.marinho@inesctec.pt",
    description="Python package containing common classes and data structures required by Tunnelis.",
    packages=setuptools.find_packages(),
    install_requires=[
        "grpcio",
        "grpcio-tools",
        "tornado",
        "zmq"
    ],
    classifiers=[
        "Programming Language :: Python :: 3"
    ],
    python_requires='>=3.6'
)



from typing import Callable, Set, Tuple
from uuid import uuid1
from zmq import IDENTITY as ZMQ_IDENTITY, STREAM as ZMQ_STREAM
from zmq.eventloop.zmqstream import ZMQStream
from zmq.sugar.context import Context

TCPMessageCallback = Callable[['TCPSocket', bytes, bytes], None]
TCPConnectionCallback = Callable[['TCPSocket', bytes], None]


class TCPSocket:
    """
    A socket for asynchronous TCP communications.

    A wrapper class around ZMQ sockets of type ZMQ_STREAM.

    """

    __message_handler: TCPMessageCallback = None
    __connection_handler: TCPConnectionCallback = None
    __disconnection_handler: TCPConnectionCallback = None

    __connected_clients: Set[bytes] = set()

    def __init__(self):
        """
        Initializes socket.

        """

        self.id = str(uuid1()).encode()

        self.__zmq_context = Context()

        self.__socket = self.__zmq_context.socket(ZMQ_STREAM)
        self.stream = ZMQStream(self.__socket)
        self.stream.on_recv(self.__message_handler)

    def bind(self, port: int) -> None:
        """
        Bind socket to a TCP address and start listening for communications.

        Parameters
        ----------
        port : int
            The network port at which to bind the socket.
            When set to a non-positive integer the socket will bind to a random free port.

        """

        if port > 0:
            self.__socket.bind(f'tcp://*:{port}')
            self.port = port
        else:
            self.port = self.__socket.bind_to_random_port('tcp://*', min_port=30000)

    def connect(self, address: Tuple[str, int]) -> None:
        """
        Connect to a TCP address.

        Parameters
        ----------
        address : Tuple(str, int)
            The TCP network address to which to connect to.

        """

        host, port = address
        self.__socket.connect(f'tcp://{host}:{port}')

        # It is only required to set an identity when in client mode.
        self.identity = self.__socket.getsockopt(ZMQ_IDENTITY)

    def send(self, client_identity: bytes, message: bytes) -> None:
        """
        Send raw application data to a remote client.

        Parameters
        ----------
        identity : bytes
            The identity (ZMQ_IDENTITY) of the remote client.

        message : bytes
            The TCP message to be sent.

        """
        if client_identity:  # server mode
            self.stream.send_multipart([client_identity, message])
        else:  # client mode
            self.stream.send_multipart([self.identity, message])

    def close(self) -> None:
        """
        Close endpoint.

        """
        self.stream.close()
        self.__socket.close()
        self.__zmq_context.term()

    def set_message_handler(self, handler: TCPMessageCallback) -> None:
        """
        Set the callback function to handle incoming TCP messages.

        Parameters
        ----------
        handler : Callable(TCPSocket, bytes, bytes) -> None
            The callback function.

        """
        self.__message_handler = handler

    def set_connection_handler(self, handler: TCPConnectionCallback) -> None:
        """
        Set the callback function to handle new connections to/from other TCP applications.

        Parameters
        ----------
        handler : Callable(TCPSocket, bytes) -> None
            The callback function.

        """
        self.__connection_handler = handler

    def set_disconnection_handler(self, handler: TCPConnectionCallback) -> None:
        """
        Set the callback function to handle disconnections from other TCP applications.

        Parameters
        ----------
        handler : Callable(TCPSocket, bytes) -> None
            The callback function.

        """
        self.__disconnection_handler = handler

    def unset_message_handler(self) -> None:
        """
        Stop handling incoming TCP messages.

        """
        self.__message_handler = None

    def unset_connection_handler(self) -> None:
        """
        Stop handling connections to/from other TCP applications.

        """
        self.__connection_handler = None

    def unset_disconnection_handler(self) -> None:
        """
        Stop handling disconnections from other TCP applications.

        """
        self.__disconnection_handler = None

    def __message_handler(self, message: Tuple[bytes, bytes]) -> None:
        """
        A generic message handler.
        Forwards received messages to their proper message handlers.

        Parameters
        ----------
        message : Tuple(bytes, bytes)
            The received message.

        """

        identity, payload = message

        if not payload:

            if identity not in self.__connected_clients:

                if identity != self.id:

                    self.__connected_clients.add(identity)
                    if self.__connection_handler:
                        self.__connection_handler(self, identity)

            else:

                self.__connected_clients.remove(identity)
                if self.__disconnection_handler:
                    self.__disconnection_handler(self, identity)

        else:

            if self.__message_handler:
                self.__message_handler(self, identity, payload)



from typing import Tuple
from zmq import PUB as ZMQ_PUB
from zmq.sugar.context import Context


class Publisher:
    """
    A generic message publisher.

    A wrapper class around ZMQ sockets of type ZMQ_PUB.

    """

    def __init__(self):
        """
        Initialize a message publisher.

        """
        self.__zmq_context = Context()

        # Connect publisher to message broker
        self.__socket = self.__zmq_context.socket(ZMQ_PUB)

    def connect(self, broker: Tuple[str, int]) -> None:
        """
        Connect publisher to an external message broker.

        Parameters
        ----------
        broker : Tuple(str, int)
            The TCP network address (host, port) of the message broker.
        """

        broker_host, broker_port = broker
        self.__socket.connect(f'tcp://{broker_host}:{broker_port}')

    def bind(self, address: Tuple[str, int]) -> None:
        """
        Bind publisher to an address.

        Parameters
        ----------
        address : Tuple(str, int)
            The TCP network address (host, port) for this publisher.
        """

        host, port = address
        self.__socket.connect(f'tcp://{host}:{port}')

    def publish(self, topic: bytes, message: bytes) -> None:
        """
        Publish a message.

        Parameters
        ----------
        topic : bytes
            The topic on which to publish the message.

        message : bytes
            The message to be published.

        """
        self.__socket.send_multipart([topic, message])

    def close(self) -> None:
        """
        Stops publisher.

        """
        self.__socket.close()
        self.__zmq_context.term()
